%mfield
% Function which creates a 3D visulization of the magnetic field strength
% resulting from a set of current elements. Default output is the field in the
% vicinity of a solenoid. Current is assumed to be steady and therefore the
% Biot-Savart equations are used for the calculation of the net magnetic field.
%
% For computational simplicity, Stokes' Theorem is employed, and the
% magnetic field is computed via the curl of the magnetic vector potential.
%
% Note if the current is not steady then the resulting B field must be computed
% using the Jefimenko equations.
% https://en.wikipedia.org/wiki/Jefimenko%27s_equations
%
% [Bx,By,Bz,x,y,z] = mfield( xI, yI,zI, Jx, Jy, Jz, I, xlimits, ylimits, zlimits plot_title )
%
% xI        Vector determining x coordinates of current element Idl
%           where dl = sqrt(dx^2 + dy^2 + dz^2)
% yI        Vector determining y coordinates of current element Idl
% zI        Vector determining z coordinates of current element Idl
% Jx        Current density in x direction /amps per m^2
% Jy        Current density in y direction /amps per m^2
% Jz        Current density in z direction /amps per m^2
% I         Uniform current /amps
% xlimits   [xmin,xmax] vector defining plot limits /m
% ylimits   [ymin,ymax] vector defining plot limits /m
% zlimits   [zmin,zmax] vector defining plot limits /m
%
% Bx        Matrix of magnetic field strength horizontal coordinates at (x,y,z)
% By        Matrix of magnetic field strength vertical  coordinates at (x,y,z)
% Bz        Matrix of magnetic field strength vertical  coordinates at (x,y,z)
% x         Matrix of x values
% y         Matrix of y values
% z         Matrix of z values
%
% LAST UPDATED by Andy French Oct 2016.

function varargout = mfield(varargin)

%Define grid size (there will be N^2 data points)
N = 100;

%Number of direction arrow evaluation points for streamlines and current loop
Nd = 10;

%Fontsize for graphs
fsize = 16;

%Permeability of free space / m kg s^-2 A^-2 or N/A^2
mu0 = 4*pi*1e-7;

%

%Assign inputs
if nargin==0
    
    %Default inputs ...
    
    %Number of current elements
    M = 500;
    
    %Length of solenoid
    L = 0.1;
    
    %Radius of solenoid
    r = L;
    
    %Wire diameter in m
    d = 1e-3;
    
    %Number of helical wire coils or radius r
    num_turns = 10;
    
    %Compute coordinates of current elements
    t=linspace(0,2*pi*num_turns,M);
    xI = r*cos(t);
    yI = linspace(-L/2,L/2,M);
    zI = r*sin(t);
    
    %Define current /square metre ('current density') at current element
    %locations
    I = 1;  %1 amp curent
    Jx = -sin(t);
    Jz = cos(t);
    Jy = zeros(1,M);
    
    %Define limits for plot
    xlimits = 3*L*[-1,1];
    ylimits = 3*L*[-1,1];
    zlimits = 3*L*[-1,1];
    plot_title = 'Solenoid';
else
    %Assume all inputs are fully specified
    xI = varargin{1};
    yI = varargin{2};
    zI = varargin{3};
    Jx = varargin{4};
    Jy = varargin{5};
    Jz = varargin{6};
    I = varargin{7};
    xlimits = varargin{8};
    ylimits = varargin{9};
    zlimits = varargin{10};
    plot_title = varargin{11};
end

%

%Define x,y grid
xx = linspace( xlimits(1), xlimits(2), N );
yy = linspace( ylimits(1), ylimits(2), N );
zz = linspace( zlimits(1), zlimits(2), N );
[x,y,z] = meshgrid(xx,yy,zz);

%Determine positions and directions for arrows to be overlaid upon the plot
%of the current loop
M = numel(xI);
xId = interp1( 1:M, xI, linspace(1,M,Nd) );
yId = interp1( 1:M, yI, linspace(1,M,Nd) );
zId = interp1( 1:M, zI, linspace(1,M,Nd) );
Jxd = interp1( 1:M, Jx, linspace(1,M,Nd) );
Jyd = interp1( 1:M, Jy, linspace(1,M,Nd) );
Jzd = interp1( 1:M, Jz, linspace(1,M,Nd) );

%Determine length of current loop and hence determine current x length
%element length (I*dl)
xxx = xI(1);
yyy = yI(1);
zzz = zI(1);
l = 0;
for k=1:M
    l = l + sqrt( (xI(k) - xxx)^2 + (yI(k) - yyy)^2 + (zI(k) - zzz)^2 );
    xxx = xI(k);
    yyy = yI(k);
    zzz = zI(k);
end
dl = l/M;

%Magnitude of current density J.
J = sqrt( Jx.^2 + Jy.^2 + Jz.^2 );

%Normalize i.e. so it just gives the direction of current element I*dl
Jx = Jx./J;
Jy = Jy./J;
Jz = Jz./J;

%Determine magnetic field for each point in the grid due to
%the current elements using the Biot-Savart law, Stokes' Theorem, Magnetic
%vector potential
Ax = zeros(N,N,N);
Ay = zeros(N,N,N);
Az = zeros(N,N,N);
o = ones(N,N,N);
for k=1:numel(xI)
    %Distance from current element to grid point
    r = sqrt( ( x - xI(k)*o ).^2 + ( y - yI(k)*o ).^2 + ( z - zI(k)*o ).^2 );
    
    %Add to x,y,z components of magnetic vector potential
    Ax  = Ax + ( mu0/(4*pi) )*I*dl*Jx(k)./r;
    Ay  = Ay + ( mu0/(4*pi) )*I*dl*Jy(k)./r;
    Az  = Az + ( mu0/(4*pi) )*I*dl*Jz(k)./r;
end

%Determine magnetic field strength
[Bx,By,Bz] = curl( x,y,z, Ax, Ay, Az );

%Determine log10 of magnetic field /Tesla strength at all grid points
B = log10( sqrt( Bx.^2 + By.^2 + Bz.^2 ) );

%Determine maximum and minimum field strengths
Bmin = min(min(min(B)));
Bmax = max(max(max(B)));

%Define outputs
if nargout~=0
    varargout{1} = Bx;
    varargout{2} = By;
    varargout{3} = Bz;
    varargout{4} = x;
    varargout{5} = y;
    varargout{6} = z;
end

%%

%% xy plane slice %%

%Determine xy slice plane
B_xy = zeros(N,N);
Bx_xy = zeros(N,N);
By_xy = zeros(N,N);
[xx,yy] = meshgrid( linspace(xlimits(1),xlimits(2),N),...
    linspace(ylimits(1),ylimits(2),N) );
B_xy = interp3( x,y,z,B, xx,yy,zeros(N,N));
Bx_xy = interp3( x,y,z,Bx, xx,yy,zeros(N,N) );
By_xy = interp3( x,y,z,By, xx,yy,zeros(N,N) );

%Plot in new figure
figure('color',[1 1 1],'name','magnetic field xy slice');
pcolor(xx,yy,B_xy);
shading interp
hold on;

%Set axis limits
axis equal;
xlim(xlimits);
zlim(zlimits);
axis manual

%Set colour scale
caxis([Bmin,Bmax]);
interp_colormap(500);
colorbar('fontsize',fsize);

%Overlay field lines in x,y plane
[verts, averts] = streamslice(xx,yy,Bx_xy,By_xy);
s = streamline([verts averts]);
set(s,'color','w');

%Overlay current elements as a continuous red line
plot( xI,yI, 'm-','linewidth',2)

%Grid lines and annotation
grid on;
set(gca,'fontsize',fsize);
xlabel('x /m','fontsize',fsize)
ylabel('y /m','fontsize',fsize)
title({plot_title,...
    'Colour scale is log_{10} of B field in Tesla'},'fontsize',fsize)

%Save a .png file and close figure
print(gcf,[plot_title,'_xz.png'],'-dpng','-r300' );
close(gcf)

%%

%% 3D Magnetic field visualization with xy plane streamline overlay %%

%Create new figure
figure('color',[1 1 1],'name','magnetic field');

%Define values to evaluate iso-surface of magnetic field strength
k = linspace(0,1,5);
BB = Bmin + k*(Bmax-Bmin);    %Field values
t = 0.1+0.9*exp(4*(k-1));     %Transparency values

%Plot magnetic field strength as a volume visualization
interp_colormap(500);
vol( x,y,z,B,BB,t,[] );
caxis([Bmin,Bmax]);
lighting phong
camlight
hold on;

%Set axis limits and add colorbar
axis equal;
xlim(xlimits);
ylim(ylimits);
ylim(zlimits);
axis manual
colorbar('fontsize',fsize);

%Overlay field lines or field vectors
s = streamline([verts averts]);
set(s,'color','b');

%Overlay current elements as a continuous red line
plot3( xI,yI,zI, 'r-','linewidth',2)

%Add current direction arrows
quiver3( xId,yId,zId, Jxd, Jyd, Jzd, 0.5, 'r');

%Grid lines and annotation
grid on;
set(gca,'fontsize',fsize);
xlabel('x /m','fontsize',fsize)
ylabel('y /m','fontsize',fsize)
zlabel('z /m','fontsize',fsize)
title({plot_title,...
    'Colour scale is log_{10} of B field in Tesla'},'fontsize',fsize)

%Save a .png file and close figure
print(gcf,[plot_title,'.png'],'-dpng','-r300' );
close(gcf)

%%

%interp_colormap
% Function which interpolates current colormap to yield better graduated
% shading. N is number of possible colours.
function interp_colormap(N)

%Get current colourmap
map = colormap;

%Initialise new colormap
new_map = ones(N,3);

%Get size of current colormap and initalise red,green,blue vectors
dim = size(map);
R = ones(1,dim(1));
G = ones(1,dim(1));
B = ones(1,dim(1));
RR = ones(1,N);
GG = ones(1,N);
BB = ones(1,N);

%Populate these with current colormap
R(:) = map(:,1);
G(:) = map(:,2);
B(:) = map(:,3);

%Interpolate to yield new colour map
x = linspace( 1, dim(1), N );
RR = interp1( 1:dim(1), R, x );
GG = interp1( 1:dim(1), G, x );
BB = interp1( 1:dim(1), B, x );
new_map(:,1) = RR(:);
new_map(:,2) = GG(:);
new_map(:,3) = BB(:);

%Set colormap to be new map
colormap( new_map );

%%

function vol( x,y,z,V,v,t,cax )

%Point to current active figure, or create a new one if one doesn't exist.
fig = gcf;

%Make figure window white and set rendering option to OpenGL. Use ZBuffer
%if your graphics card doesn't support this.
set(fig,'color',[1 1 1],'renderer','opengl');

%Obtain maximum and minimum values of V
Vmax = max(max(max(V)));
Vmin = min(min(min(V)));

%Make sure v is within the range Vmin to Vmax
v(v>Vmax) = Vmax;
v(v<Vmin) = Vmin;

%Make sure t is within the range 0 to 1
t(t>1) = 1;
t(t<0) = 0;

%Set limits of V to correspond to colormap
if isempty(cax)==1
    caxis( [Vmin,Vmax] );
else
    caxis( cax );
end

%Visualize volume by plotting nested shells of volume iso-surfaces with
%varying transparency. Use current colormap to set the colour of the shells
for n=1:length(v)
    [R,G,B] = colfromcolormap( v(n) );
    p = patch( isosurface(x,y,z,V, v(n) ),...
        'Facealpha',t(n),'FaceColor', [R,G,B], 'EdgeColor', 'none');
    isonormals(x,y,z,V,p);
end
view(3);
daspect([1 1 1]);

%%

%colfromcolormap
% Obtain colours corresponding to value x from curent colormap.

function [R,G,B] = colfromcolormap(x)

%Get vaiable range of current colormap
[xmin,xmax] = caxis;

%Get current colormap
map = colormap;
dim = size(map);

%Make sure x is within the caxis range
x(x>xmax) = xmax;
x(x<xmin) = xmin;

%Interpolate R,G,B values at x
R = interp1( linspace(xmin,xmax,dim(1)),map(:,1).',x );
G = interp1( linspace(xmin,xmax,dim(1)),map(:,2).',x );
B = interp1( linspace(xmin,xmax,dim(1)),map(:,3).',x );

%End of code